"""Test I2C functionality on a FreeWili."""

import pytest

from freewili import FreeWili
from freewili.framing import ResponseFrameType


@pytest.mark.skipif("len(FreeWili.find_all()) == 0")
def test_hw_board_leds() -> None:
    """Test LEDs on a FreeWili."""
    device = FreeWili.find_all()[0]
    device.stay_open = True

    try:
        for led_num in range(7):
            response_frame = device.set_board_leds(led_num, 50, 50, 50).expect("Failed to set LED")
            assert response_frame.rf_type == ResponseFrameType.Standard
            assert response_frame.rf_type_data == r"g\s"
            assert response_frame.timestamp != 0
            assert response_frame.response == "Ok"
            assert response_frame.is_ok()

        for led_num in range(7):
            response_frame = device.set_board_leds(led_num, 0, 0, 0).expect("Failed to set LED")
            assert response_frame.rf_type == ResponseFrameType.Standard
            # assert response_frame.rf_type_data == r"g\s"
            assert response_frame.timestamp != 0
            assert response_frame.response == "Ok"
            assert response_frame.is_ok()
    finally:
        device.close()


@pytest.mark.skipif("len(FreeWili.find_all()) == 0")
def test_hw_show_gui_image() -> None:
    """Test image on a FreeWili."""
    device = FreeWili.find_all()[0]

    try:
        device.open().expect("Failed to open")
        success = device.send_file("tests/assets/pip_boy.fwi").expect("Failed to upload file")
        assert success != ""
        response_frame = device.show_gui_image("pip_boy.fwi").expect("Failed to show image")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"g\l"
        assert response_frame.timestamp != 0
        assert response_frame.response == "Ok", "Is pip_boy.fwi uploaded to the device?"
        assert response_frame.is_ok()

        response_frame = device.reset_display().expect("Failed to reset display")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"g\t"
        assert response_frame.timestamp != 0
        assert response_frame.response == "Ok"
        assert response_frame.is_ok()
    finally:
        device.close()


@pytest.mark.skipif("len(FreeWili.find_all()) == 0")
def test_hw_show_text_display() -> None:
    """Test show text on a FreeWili."""
    device = FreeWili.find_all()[0]
    device.stay_open = True

    try:
        response_frame = device.show_text_display("test").expect("Failed to show image")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"g\p"
        assert response_frame.timestamp != 0
        assert response_frame.response == "Ok"
        assert response_frame.is_ok()
    finally:
        device.close()


@pytest.mark.skipif("len(FreeWili.find_all()) == 0")
def test_hw_read_all_buttons() -> None:
    """Test read buttons on a FreeWili."""
    device = FreeWili.find_all()[0]
    device.stay_open = True

    try:
        response_frame = device.read_all_buttons().expect("Failed to show image")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"g\u"
        assert response_frame.timestamp != 0
        assert response_frame.response == "0 0 0 0 0"
        assert response_frame.is_ok()
    finally:
        device.close()


if __name__ == "__main__":
    import pytest

    pytest.main(
        args=[
            __file__,
            "--verbose",
        ]
    )



"""Example I2C on a FreeWili with SparkFun 9DoF IMU Breakout - ISM330DHCX, MMC5983MA (Qwiic) attached.

https://www.sparkfun.com/sparkfun-9dof-imu-breakout-ism330dhcx-mmc5983ma-qwiic.html
ISM330DHCX I2C Address: 0x6B (Default)
MMC5983MA Magnetometer I2C Address: 0x30
"""

import enum

from result import Err, Ok

from freewili import FreeWili

MMC5983MA_ADDR = 0x30
ISM330DHCX_ADDR = 0x6B


class MMC5983MA(enum.Enum):
    """Register list for MMC5983MA Magnetometer."""

    Xout0 = 0x00  # Xout [17:10]
    Xout1 = 0x01  # Xout [9:2]
    Yout0 = 0x02  # Yout [17:10]
    Yout1 = 0x03  # Yout [9:2]
    Zout0 = 0x04  # Zout [17:10]
    Zout1 = 0x05  # Zout [9:2]
    XYZout2 = 0x06  # Xout[1:0], Yout[1:0], Zout[1:0]
    Tout = 0x07  # Temperature output
    Status = 0x08  # Device status
    Control0 = 0x09  # Control register 0
    Control1 = 0x0A  # Control register 1
    Control2 = 0x0B  # Control register 2
    Control3 = 0x0C  # Control register 3
    ProductID = 0x2F  # Product ID


class ISM330DHCX(enum.Enum):
    """Register list for ISM330DHCX."""

    FUNC_CFG_ACCESS = 0x01
    PIN_CTRL = 0x02
    FIFO_CTRL1 = 0x07
    FIFO_CTRL2 = 0x08
    FIFO_CTRL3 = 0x09
    FIFO_CTRL4 = 0x0A
    COUNTER_BDR_REG1 = 0x0B
    COUNTER_BDR_REG2 = 0x0C
    INT1_CTRL = 0x0D
    INT2_CTRL = 0x0E
    WHO_AM_I = 0x0F
    CTRL1_XL = 0x10
    CTRL2_G = 0x11
    CTRL3_C = 0x12
    CTRL4_C = 0x13
    CTRL5_C = 0x14
    CTRL6_C = 0x15
    CTRL7_G = 0x16
    CTRL8_XL = 0x17
    CTRL9_XL = 0x18
    CTRL10_C = 0x19
    ALL_INT_SRC = 0x1
    WAKE_UP_SRC = 0x1B
    TAP_SRC = 0x1C
    DRD_SRC = 0x1D
    STATUS_REG = 0x1E
    STATUS_SPIAux = 0x1E
    OUT_TEMP_L = 0x20
    OUT_TEMP_H = 0x21
    OUTX_L_G = 0x22
    OUTX_H_G = 0x23
    OUTY_L_G = 0x24
    OUTY_H_G = 0x25
    OUTZ_L_G = 0x26
    OUTZ_H_G = 0x27
    OUTX_L_A = 0x28
    OUTX_H_A = 0x29
    OUTY_L_A = 0x2A
    OUTY_H_A = 0x2B
    OUTZ_L_A = 0x2C
    OUTZ_H_A = 0x2D
    EMB_FUNC_STATUS_MAINPAGE = 0x35
    FSM_STATUS_A_MAINPAGE = 0x36
    FSM_STATUS_B_MAINPAGE = 0x37
    MLC_STATUS_MAINPAGE = 0x38
    STATUS_MASTER_MAINPAGE = 0x39
    FIFO_STATUS1 = 0x3A
    FIFO_STATUS2 = 0x3B
    TIMESTAMP0 = 0x40
    TIMESTAMP1 = 0x41
    TIMESTAMP2 = 0x42
    TIMESTAMP3 = 0x43
    TAP_CFG0 = 0x56
    TAP_CFG1 = 0x57
    TAP_CFG2 = 0x58
    TAP_THS_6D = 0x59
    INT_DUR2 = 0x5A
    WAKE_UP_THS = 0x5B
    WAKE_UP_DUR = 0x5C
    FREE_FALL = 0x5D
    MD1_CFG = 0x5E
    MD2_CFG = 0x5F
    INTERNAL_FREQ_FINE = 0x63
    INT_OIS = 0x6F
    CTRL1_OIS = 0x70
    CTRL2_OIS = 0x71
    CTRL3_OIS = 0x72
    X_OFS_USR = 0x73
    Y_OFS_USR = 0x74
    Z_OFS_USR = 0x75
    FIFO_DATA_OUT_TAG = 0x78
    FIFO_DATA_OUT_X_L = 0x79
    FIFO_DATA_OUT_X_H = 0x7A
    FIFO_DATA_OUT_Y_L = 0x7B
    FIFO_DATA_OUT_Y_H = 0x7C
    FIFO_DATA_OUT_Z_L = 0x7D
    FIFO_DATA_OUT_Z_H = 0x7E


def get_mmc5983ma_temperature(device: FreeWili) -> tuple[float, int]:
    """Get temperature on MMC5983MA.

    Arguments:
    ----------
        device: FreeWili
            FreeWili device to use.

    Returns:
    --------
        tuple[float, int]:
            (Temperature in degrees Celcius, timestamp in ns).
    """
    resp = device.write_i2c(MMC5983MA_ADDR, MMC5983MA.Control0.value, bytes([0x2A])).expect("Enable temp measure")
    # print(resp.is_ok())
    # print("Waiting for temperature measurement...")
    while True:
        resp = device.read_i2c(MMC5983MA_ADDR, MMC5983MA.Status.value, 1).expect("Reading Status failed")
        status = resp.response_as_bytes().expect("Status bytes")[0]
        if (status & 0x2) == 0x2:
            # print("Temperature measurement done!")
            break
    resp = device.read_i2c(MMC5983MA_ADDR, MMC5983MA.Tout.value, 1).expect("Reading Tout failed")
    # Temperature output, unsigned format. The range is -75~125°C, about 0.8°C/LSB, 00000000 stands for -75°C
    temp = resp.response_as_bytes().expect("MMC5983MA.Tout")
    temp_int = int.from_bytes(temp, byteorder="little", signed=False)
    converted = temp_int * 0.8 - 75
    # print(resp._raw.strip())
    return (converted, resp.timestamp)


def get_mmc5983ma_magnetic_sensor(device: FreeWili) -> tuple[tuple[int, int, int], int]:
    """Get magnetic field data on MMC5983MA.

    Arguments:
    ----------
        device: FreeWili
            FreeWili device to use.

    Returns:
    --------
        tuple[tuple[int, int, int], int]:
            ((x, y, z), timestamp in ns).
    """
    resp = device.write_i2c(MMC5983MA_ADDR, MMC5983MA.Control0.value, bytes([0x29])).expect("Enable magnetic field")
    # print(resp.is_ok())
    # print("Waiting for temperature measurement...")
    while True:
        resp = device.read_i2c(MMC5983MA_ADDR, MMC5983MA.Status.value, 1).expect("Reading Status failed")
        status = resp.response_as_bytes().expect("Status bytes")[0]
        if (status & 0x1) == 0x1:
            # print("Temperature measurement done!")
            break
    resp = device.read_i2c(MMC5983MA_ADDR, MMC5983MA.Xout0.value, 6).expect("Reading Tout failed")
    data = resp.response_as_bytes().expect("MMC5983MA.Xout0")
    x = int.from_bytes(data[0:2], byteorder="little", signed=False) << 2
    y = int.from_bytes(data[2:4], byteorder="little", signed=False) << 2
    z = int.from_bytes(data[4:6], byteorder="little", signed=False) << 2
    # print(resp._raw.strip())
    return ((x, y, z), resp.timestamp)


# find a FreeWili device
match FreeWili.find_first():
    case Ok(d):
        device = d
        device.stay_open = True
        print(f"Using {device}")
    case Err(msg):
        print(msg)
        exit(1)

try:
    # Poll the I2C to make sure we can read the breakout board
    print("Polling I2C...")
    resp = device.poll_i2c().expect("Failed to poll i2c")
    addresses = resp.response_as_bytes().expect("Failed to get response bytes")
    if addresses != bytes([2, MMC5983MA_ADDR, ISM330DHCX_ADDR]):
        print(f"Expected addresses on I2C are incorrect, got {list(addresses)}!")
        exit(1)

    first_ts = resp.timestamp
    print(resp.timestamp_as_datetime().expect("Failed to convert to datetime"))
    while True:
        try:
            temp_c, ts = get_mmc5983ma_temperature(device)
            temp_f = temp_c * 1.8 + 32
            print(f"[{ts - first_ts}] Temperature: {temp_c:.2f}C ({temp_f:.2f}F)")
            magnetic_data, ts = get_mmc5983ma_magnetic_sensor(device)
            print(f"[{ts - first_ts}] Magnetic Field: {magnetic_data}")
        except KeyboardInterrupt:
            break
finally:
    device.close()

print("Done.")



"""Serial Port Reader/Writer."""

import queue
import threading
import time
from queue import Queue
from typing import Any

from result import Err, Ok, Result
from serial import Serial, SerialException

from freewili.framing import ResponseFrame
from freewili.util.fifo import SafeIOFIFOBuffer


class SerialPort(threading.Thread):
    """Read/Write data to a serial port."""

    def __init__(self, port: str, baudrate: int = 1000000, name: str = ""):
        self._debug_enabled = False
        self._name = name
        super().__init__(daemon=True, name=f"Thread-SerialPort-{port}-{name}")
        self._port = port
        self._baudrate = baudrate
        self._running = threading.Event()
        self._running.set()
        self._connect = threading.Event()
        self._is_connected: bool = False
        self._in_error = threading.Event()
        self._error_msg: str = ""

        self.send_queue: Queue = Queue()
        # Response frame queue
        self.rf_queue: Queue = Queue()
        self.rf_event_queue: Queue = Queue()
        # data other than a response frame
        self.data_queue: Queue = Queue()

        self.start()

    def shutdown(self) -> None:
        """Shutdown the reader."""
        self._running.clear()
        self.join()

    def open(self, block: bool = True, timeout_sec: float = 6.0) -> Result[None, str]:
        """Open the serial port.

        See also: is_open()

        Parameters:
        ----------
            block: bool:
                If True, block until the serial port is opened.
            timeout_sec: float:
                number of seconds to wait when blocking.

        Returns:
        -------
            None

        Raises:
        ------
            TimeoutError:
                When blocking is True and time elapsed is greater than timeout_sec
        """
        assert isinstance(block, bool)
        assert isinstance(timeout_sec, float)
        self._connect.set()
        if block:
            start = time.time()
            while time.time() - start < timeout_sec and not self.is_open():
                if self.has_error():
                    break
                time.sleep(0.001)
            if not self.is_open():
                return Err(f"Failed to open in {timeout_sec:.1f} seconds: {self.get_error()}")
        else:
            return Ok(None)
        if not self.is_open():
            return Err(f"Failed to open serial Port: {self.get_error()}")
        return Ok(None)

    def close(self, block: bool = True, timeout_sec: float = 6.0) -> None:
        """Close the serial port.

        See also: is_open()

        Parameters:
        ----------
            block: bool:
                If True, block until the serial port is closed.
            timeout_sec: float:
                number of seconds to wait when blocking.

        Returns:
        -------
            None

        Raises:
        ------
            TimeoutError:
                When blocking is True and time elapsed is greater than timeout_sec
        """
        assert isinstance(block, bool)
        assert isinstance(timeout_sec, float)
        self._connect.clear()
        if block:
            start = time.time()
            current = time.time()
            while current - start < timeout_sec and self.is_open():
                current = time.time()
                time.sleep(0.001)
            if self.is_open():
                raise TimeoutError(f"Failed to close serial port in {timeout_sec:.1f} seconds.")

    def is_open(self) -> bool:
        """Return if the serial port is open.

        Parameters:
        ----------
            None

        Returns:
        -------
            bool:
                True if open, False if closed.
        """
        return self._is_connected

    def has_error(self) -> bool:
        """Return if the serial port is in an error state.

        To clear the error state, call get_error().

        Parameters:
        ----------
            None

        Returns:
        -------
            bool:
                True if there are errors, False otherwise.
        """
        return self._in_error.is_set()

    def get_error(self) -> str:
        """Get the serial port error message. Clears the error state.

        Parameters:
        ----------
            None

        Returns:
        -------
            str:
                Error message if present, empty str otherwise.
        """
        if not self.has_error():
            return ""
        msg = self._error_msg
        self._in_error.clear()
        return msg

    @property
    def port(self) -> str:
        """Get the serial port descriptor.

        Returns:
        -------
            str:
                serial port descriptor.
        """
        return self._port

    @property
    def baudrate(self) -> int:
        """Get the serial port baudrate.

        Returns:
        -------
            str:
                serial port baudrate.
        """
        return self._baudrate

    def run(self) -> None:
        """Thread handler function. Call Self.start() to initialize."""
        self._debug_print(f"Started {self._port}...\n")
        serial_port: None | Serial = None
        # read_buffer_data: bytearray = bytearray()
        # read_buffer = io.BytesIO()
        read_buffer = SafeIOFIFOBuffer(blocking=False)
        start_time = time.time()
        while self._running.is_set():
            if self._in_error.is_set():
                time.sleep(0.001)
                continue
            try:
                # Configure the serial port
                if self._connect.is_set():
                    # We are allowed to connect
                    if not serial_port:
                        try:
                            self._debug_print(f"[{time.time() - start_time:.3f}] Opening {self._port}...\n")
                            serial_port = Serial(
                                self._port,
                                baudrate=1000000,
                                timeout=0.001,
                                exclusive=True,
                                rtscts=False,
                                xonxoff=False,
                                dsrdtr=False,
                            )
                            # This is absolutely needed, for some reason writing data too fast after open
                            # will corrupt things and the read buffer does strange things.
                            # 0.1 was successful 50% of the time in my testing and 0.2 was 100% successful.
                            # 0.5 should allow for other slower systems if its a timing issue on the OS kernel level?
                            time.sleep(0.5)
                            self._is_connected = True
                        except SerialException as ex:
                            print(ex)
                            self._error_msg = str(ex)
                            self._in_error.set()
                            continue
                else:
                    # We are allowed to disconnect
                    if serial_port and self.send_queue.empty():
                        self._debug_print(f"[{time.time() - start_time:.3f}] Closing {self._port}...\n")
                        serial_port.close()
                        serial_port = None
                        self._is_connected = False
                        continue
                    elif serial_port and not self.send_queue.empty():
                        self._debug_print(
                            f"[{time.time() - start_time:.3f}] Send queue not empty yet, waiting to close port...\n"
                        )
                    else:
                        # serial_port isn't valid here, tight loop back to the beginning.
                        time.sleep(0.001)
                        continue
                # Send data
                try:
                    send_data, delay_sec = self.send_queue.get_nowait()
                    # self._debug_print(f"[{time.time() - start_time:.3f}] sending: ", send_data, self._port)
                    write_len = serial_port.write(send_data)
                    # self._debug_print(f"[{time.time() - start_time:.3f}]: Delaying for {delay_sec:.3f} seconds...")
                    time.sleep(delay_sec)
                    self.send_queue.task_done()
                    if len(send_data) != write_len:
                        self._debug_print("[{time.time()-start_time:.3f}] ERROR: send_data != write_len")
                    assert len(send_data) == write_len, f"{len(send_data)} != {write_len}"
                except queue.Empty:
                    pass
                # Read data
                if serial_port.in_waiting > 0:
                    # self._debug_print(f"[{time.time() - start_time:.3f}] Reading {serial_port.in_waiting}...")
                    data = serial_port.read(4096)
                    if data != b"":
                        read_buffer.write(data)
                        # self._debug_print(f"[{time.time() - start_time:.3f}] RX: ", repr(data), len(data))
                    # self._debug_print("handle data...")
                self._handle_data(read_buffer)
            except Exception as ex:
                self._error_msg = str(ex)
                self._debug_print(f"Exception: {type(ex)}: {self._error_msg}")
                self._in_error.set()
                if serial_port and serial_port.is_open:
                    serial_port.close()
                    serial_port = None
        if serial_port:
            serial_port.close()
        self._debug_print("Done.")

    def _debug_print(self, *args: Any, **kwargs: Any) -> None:
        if self._debug_enabled:
            print(*args, **kwargs)

    _debug_count: int = 0

    def _handle_data(self, data_buffer: SafeIOFIFOBuffer) -> None:
        assert isinstance(data_buffer, SafeIOFIFOBuffer)
        if data_buffer.available() == 0:
            return
        # Match a full event response frame
        while frame := data_buffer.pop_first_match(rb"\[\*.*.\d\]\r?\n"):
            self._debug_print(f"RX Event Frame: {frame!r}")
            # self._debug_print(f"Buffer len: {data_buffer.available()} {data_buffer.peek()!r}")
            self.rf_event_queue.put(ResponseFrame.from_raw(frame))
            _debug_count = 0
        # Match a full response frame
        while frame := data_buffer.pop_first_match(rb"\[[^\*].*.\d\]\r?\n"):
            self._debug_print(f"RX Frame: {frame!r}")
            # self._debug_print(f"Buffer len: {data_buffer.available()} {data_buffer.peek()!r}")
            self.rf_queue.put(ResponseFrame.from_raw(frame))
            self._debug_count = 0
        # Match anything else
        try:
            # add anything before [ to the data queue
            start, end = data_buffer.contains(rb"\[")
            if start > 0:
                data = data_buffer.read(start)
                self._debug_print(f"RX Data: {len(data)}: {self._debug_count}: {data!r}")
                self.data_queue.put(data)
                self._debug_count += len(data)
        except ValueError:
            pass

        # At this point we should be at the start of a frame
        data_len = data_buffer.available()
        data = data_buffer.peek(data_len)
        # If we only have a single byte and it's a [, we are done
        if data_len == 1 and data == b"[":
            return
        if data_len == 2 and data == b"[*":
            return
        if data_len >= 3:
            try:
                # [*f or [a\
                _start, _end = data_buffer.contains(rb"(\[\*)|(\[.\\)|(\[. )")
                # This is probably a start of a frame, do nothing
                return
            except ValueError:
                pass
        data = data_buffer.read(-1)
        self._debug_print(f"RX Data: {len(data)}: {self._debug_count}: {data!r}")
        self.data_queue.put(data)
        self._debug_count += len(data)

    def send(
        self,
        data: bytes | str,
        append_newline: bool = True,
        newline_chars: str = "\n",
        delay_sec: float = 0.001,
        wait: bool = True,
    ) -> None:
        r"""Send data to the serial port.

        Parameters:
        ----------
            data : bytes | str:
                data to be sent to the serial port. If type is str it will be automatically encoded.
            append_newline : bool:
                Appends "\r\n" to the data if True.
            newline_chars : str:
                Appends to data if append_newline is True.
            delay_sec : float:
                Number of seconds to wait after sending.

        Returns:
        -------
            None
        """
        assert isinstance(data, (bytes, str))
        if isinstance(data, str):
            data = data.encode("ascii")
        if append_newline:
            data += newline_chars.encode("ascii")
        self._debug_print("send:", data, delay_sec)
        self.send_queue.put((data, delay_sec))
        if wait:
            self.send_queue.join()

    def clear(self) -> None:
        """Clear all the data in the queues."""
        queues = (self.rf_queue, self.data_queue)
        for q in queues:
            try:
                while True:
                    q.get_nowait()
            except queue.Empty:
                pass



"""Set Board RGB on a Free-Wili."""

import time

from result import Err, Ok

from freewili import FreeWili
from freewili.framing import ResponseFrame


def check_resp(device: FreeWili, resp: ResponseFrame) -> None:
    """Check the response and exit on error."""
    if not resp.is_ok():
        print(f"Response failed: {resp.response}")
        device.close()
        exit(1)


# find a FreeWili device
match FreeWili.find_first():
    case Ok(d):
        device = d
        device.stay_open = True
        print(f"Using {device}")
    case Err(msg):
        print(msg)
        exit(1)

try:
    # Turn the LEDs on
    for led_num in range(7):
        resp = device.set_board_leds(led_num, 10, 10, led_num * 2).expect("Failed to set LED")
        check_resp(device, resp)
    # Wait so we can see them
    time.sleep(3)
    # Turn the LEDS off
    for led_num in reversed(range(7)):
        resp = device.set_board_leds(led_num, 0, 0, 0).expect("Failed to set LED")
        check_resp(device, resp)

finally:
    device.close()
